/*
    Valtra Tractors
    Copyright 2004 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Martin Storsj
    martin@martin.st
*/

#include <SDL.h>
#include <SDL_image.h>
#include <SDL_ttf.h>
#include <string.h>
#include <math.h>
#include "dialog.h"
#include "game.h"
#include "effects.h"
#include "fsys.h"

#ifndef M_PI
#define M_PI 3.141592654
#endif

const char *TextArea::strcopy(const char *text) {
	int l = strlen(text)+1;
	char *ptr = new char[l];
	strcpy(ptr,text);
	return ptr;
}

TextArea::TextArea(TTF_Font *f, const char *text, int width, IPoint p, SDL_Color col, bool c) {
	font = f;
	pos = p;
	color = col;
	centered = c;

	char *buffer = new char[strlen(text)+1];
	int lastPos = -1;
	int offset = 0;
	while (true) {
		int w, h;
		const char *inPtr = text+offset;
		char *outPtr = buffer;
		if (lastPos>-1) {
			inPtr += lastPos;
			outPtr += lastPos;
			*outPtr++ = *inPtr++;
		}
		while (*inPtr!=' ' && *inPtr!='\n' && *inPtr!='\0') *outPtr++ = *inPtr++;
		*outPtr = '\0';
		TTF_SizeText(font,buffer,&w,&h);
		if (w>=width || *inPtr=='\n' || *inPtr=='\0') {
			if (lastPos>-1 && w>=width)
				buffer[lastPos] = '\0';
			strings.push_back(strcopy(buffer));
			if (lastPos>-1 && w>=width) {
				offset += lastPos+1;
			} else {
				offset += (outPtr-buffer)/sizeof(char)+1;
				if (*inPtr=='\0') break;
			}
			lastPos = -1;

		} else {
			lastPos = (outPtr-buffer)/sizeof(char);
		}
	}
	delete [] buffer;
}

TextArea::~TextArea() {
	for (unsigned int i = 0; i < strings.size(); i++)
		delete [] strings[i];
}


void TextArea::setPos(IPoint p) {
	pos = p;
}

void TextArea::getIds(SpriteEngine *se) {
	ids.clear();
	for (unsigned int i = 0; i < strings.size(); i++)
		ids.push_back(se->getNextId());
}

void TextArea::addSprites(SpriteEngine *se) {
	SDL_Rect rect;
	rect.y = pos.y;
	for (unsigned int i = 0; i < strings.size(); i++) {
		int w, h;
		TTF_SizeText(font,strings[i],&w,&h);
		rect.x = pos.x;
		rect.w = w;
		rect.h = h;
		if (centered) rect.x -= w/2;
		se->addSprite(ids[i],new Sprite(NULL,rect,false));
		rect.y += TTF_FontHeight(font);
	}
}

void TextArea::updateSprites(SpriteEngine *se) {
	for (unsigned int i = 0; i < strings.size(); i++) {
		Sprite *s = se->getSprite(ids[i]);
		if (s==NULL) continue;
		if (!s->update) continue;
		s->img = TTF_RenderText_Blended(font,strings[i],color);
	}
}

void TextArea::doneDrawing(SpriteEngine *se) {
	for (unsigned int i = 0; i < strings.size(); i++) {
		Sprite *s = se->getSprite(ids[i]);
		if (s==NULL) continue;
		if (s->img) SDL_FreeSurface(s->img);
	}
}


Dialog::Dialog(DialogType t, int d) {
	type = t;
	data = d;
//	dialogBkg = IMG_Load("dialog.png");
	dialogBkg = NULL;
	switch (type) {
	case PLAYERS:
		dialogBkg = IMG_Load("playerbkg.png");
		break;
	case QUIT:
	case NEWGAME:
		dialogBkg = IMG_Load("yesnobkg.png");
		break;
	case CHOOSE:
		dialogBkg = IMG_Load("choose.png");
		break;
	case RESULT:
		dialogBkg = IMG_Load("result.png");
		break;
	}
	surf = SDL_CreateRGBSurface(SDL_HWSURFACE,dialogBkg->w,dialogBkg->h,32,dialogBkg->format->Rmask,dialogBkg->format->Gmask,dialogBkg->format->Bmask,dialogBkg->format->Amask);
//	surf = SDL_ConvertSurface(dialogBkg,dialogBkg->format,SDL_SWSURFACE);
	SDL_FillRect(surf,NULL,SDL_MapRGBA(surf->format,0,0,0,0));
	SDL_SetAlpha(dialogBkg,SDL_RLEACCEL,0); // don't blend the bkg into the surface, just copy the alpha values

/*	for (int i = 0; i < 3; i++) {
		char buffer[100];
		sprintf(buffer,"dialog-button%d.png",i+1);
		buttonSurf[i] = IMG_Load(buffer);
		if (type==PLAYERS) {
			sprintf(buffer,"dialog-arrow%d.png",i+1);
			arrow[i] = IMG_Load(buffer);
		} else {
			arrow[i] = NULL;
		}
	}

	font2 = TTF_OpenFont("Vera.ttf",40);
	TTF_SetFontStyle(font2,TTF_STYLE_BOLD);
*/

	wrapper = NULL;

	buttonover = buttondown = -1;
	quit = false;

	initTTF = false;
	font = NULL;

	if (type==QUIT) {
		SDL_Surface *tmp = IMG_Load("quit.png");
		SDL_Rect pos = { 27,16,0,0 };
		SDL_BlitSurface(tmp,NULL,dialogBkg,&pos);
		SDL_FreeSurface(tmp);
	} else if (type==NEWGAME) {
		SDL_Surface *tmp = IMG_Load("newgame.png");
		SDL_Rect pos = { 28,16,0,0 };
		SDL_BlitSurface(tmp,NULL,dialogBkg,&pos);
		SDL_FreeSurface(tmp);
	} else if (type==RESULT) {
		initTTF = !TTF_WasInit();
		if (initTTF) TTF_Init();
		font = TTF_OpenFont("Vera.ttf",16);
	}

	for (int i = 0; i < 4; i++)
		tractor[i] = NULL;

	const char *str = NULL;
	switch (type) {
/*	case QUIT:
		str = "Do you really want to quit?";
		buttons.push_back(new Button());
		buttons[0]->setSurf(buttonSurf[0],buttonSurf[1],buttonSurf[2]);
		buttons[0]->setText("Yes",25,20,font2);
		buttons[0]->setClickOffset(2,2);
		buttons[0]->setPos(55,220);
		buttons.push_back(new Button());
		buttons[1]->setSurf(buttonSurf[0],buttonSurf[1],buttonSurf[2]);
		buttons[1]->setText("No",30,20,font2);
		buttons[1]->setClickOffset(2,2);
		buttons[1]->setPos(220,220);
		break;
*/
	case QUIT:
	case NEWGAME:
		for (int i = 0; i < 2; i++) {
			char buffer1[25], buffer2[25], buffer3[25];
			sprintf(buffer1,"yesno%d-1.png",i);
			sprintf(buffer2,"yesno%d-2.png",i);
			sprintf(buffer3,"yesno%d-3.png",i);
			buttons.push_back(new Button());
			buttons[i]->setSurf(IMG_Load(buffer1),IMG_Load(buffer2),IMG_Load(buffer3));
			buttons[i]->setPos(70+70*i,55);
		}
		break;
/*
	case NEWGAME:
		str = "Do you want to play again?";
		buttons.push_back(new Button());
		buttons[0]->setSurf(buttonSurf[0],buttonSurf[1],buttonSurf[2]);
		buttons[0]->setText("Yes",25,20,font2);
		buttons[0]->setClickOffset(2,2);
		buttons[0]->setPos(55,220);
		buttons.push_back(new Button());
		buttons[1]->setSurf(buttonSurf[0],buttonSurf[1],buttonSurf[2]);
		buttons[1]->setText("No",30,20,font2);
		buttons[1]->setClickOffset(2,2);
		buttons[1]->setPos(220,220);
		break;
*/
	case PLAYERS:
/*		str = "How many players?";
		buttons.push_back(new Button());
		buttons[0]->setSurf(buttonSurf[0],buttonSurf[1],buttonSurf[2]);
		buttons[0]->setText("Cancel",10,20,font2);
		buttons[0]->setClickOffset(2,2);
		buttons[0]->setPos(190,220);
		for (int i = 1; i < 4; i++) {
			char str[5];
			sprintf(str,"%d",i+1);
			buttons.push_back(new Button());
			buttons[i]->setSurf(buttonSurf[0],buttonSurf[1],buttonSurf[2]);
			buttons[i]->setText(str,40,20,font2);
			buttons[i]->setClickOffset(2,2);
			buttons[i]->setPos(55+((i%2==0)?135:0),130+(((i-1)/2==1)?90:0));
		}
*/
		for (int i = 0; i <= 4; i++) {
			char buffer1[25], buffer2[25], buffer3[25];
			sprintf(buffer1,"player%d-1.png",i);
			sprintf(buffer2,"player%d-2.png",i);
			sprintf(buffer3,"player%d-3.png",i);
			buttons.push_back(new Button());
			buttons[i]->setSurf(IMG_Load(buffer1),IMG_Load(buffer2),IMG_Load(buffer3));
			if (i==0) buttons[i]->setPos(55,195);
			else buttons[i]->setPos(28,25+40*(i-1));
		}
		break;
	case CHOOSE:
		for (int i = 0; i < 4; i++) {
			char buffer[100];
			sprintf(buffer,"tractor%d-big.png",i+1);
			tractor[i] = IMG_Load(buffer);
		}
		result = 0;
		current = 2;
		buttons.push_back(new Button());
		buttons[0]->setSurf(IMG_Load("ok1.png"),IMG_Load("ok2.png"),IMG_Load("ok3.png"));
		buttons[0]->setPos(130,170);
		buttons.push_back(new Button());
		buttons[1]->setSurf(IMG_Load("arrow1-1.png"),NULL,IMG_Load("arrow1-3.png"));
		buttons[1]->setPos(35,85);
		buttons.push_back(new Button());
		buttons[2]->setSurf(IMG_Load("arrow2-1.png"),NULL,IMG_Load("arrow2-3.png"));
		buttons[2]->setPos(235,85);
		break;

	case RESULT:
		buttons.push_back(new Button());
		buttons[0]->setSurf(IMG_Load("ok1.png"),IMG_Load("ok2.png"),IMG_Load("ok3.png"));
		buttons[0]->setPos(70,240);
		for (int i = 0; i < 4; i++) {
			int score = (data>>(8*i))&0xFF;
			if (score > 0) {
				char buffer[100];
				sprintf(buffer,"tractor%d.png",i+1);
				SDL_Surface *tmp = IMG_Load(buffer);
				SDL_PixelFormat fmt = ARGB_fmt;
				SDL_Surface *tmp2 = SDL_ConvertSurface(tmp,&fmt,SDL_SWSURFACE);
				tractor[i] = Effects::rotate(tmp2,3*2*M_PI/4,true);
				SDL_FreeSurface(tmp);
				SDL_FreeSurface(tmp2);
				for (unsigned int j = 0; j <= stats.size(); j++) {
					if (j == stats.size() || score>stats[j].second) {
						stats.insert(stats.begin()+j,pair<int,int>(i,score));
						break;
					}
				}
			}
		}
		for (unsigned int i = 0; i < stats.size(); i++) {
			char buffer[10];
			sprintf(buffer,"%d.",i+1);
			SDL_Color col = { 0x00, 0x00, 0x00 };
			if (i==0 || stats[i-1].second>stats[i].second)
				texts.push_back(new TextArea(font,buffer,1000,IPoint(29,50+49*i),col));
			sprintf(buffer,"%d p",stats[i].second);
			texts.push_back(new TextArea(font,buffer,1000,IPoint(55,50+49*i),col));
		}
		break;

	default:
		str = NULL;
		break;
	}

	switch (type) {
	case QUIT:
	case NEWGAME:
	case PLAYERS:
	case CHOOSE:
	case RESULT:
		for (unsigned int i = 0; i < buttons.size(); i++) {
			buttons[i]->freeSurfaces = true;
			buttons[i]->imgOffset[2] = IPoint(3,1);
		}
		break;
	default:
		break;
	}

	text = NULL;
//	SDL_Color col = { 0x00, 0x00, 0x00 };
//	if (str) text = new TextArea(font1,str,dialogBkg->w-80,IPoint(40,40),col);
}

Dialog::~Dialog() {
	SDL_FreeSurface(dialogBkg);
	SDL_FreeSurface(surf);
	for (unsigned int i = 0; i < buttons.size(); i++)
		delete buttons[i];
	for (unsigned int i = 0; i < texts.size(); i++)
		delete texts[i];
	for (int i = 0; i < 4; i++)
		SDL_FreeSurface(tractor[i]);
/*
	for (int i = 0; i < 3; i++) {
		SDL_FreeSurface(buttonSurf[i]);
		SDL_FreeSurface(arrow[i]);
	}
	delete text;
	TTF_CloseFont(font1);
*/
	if (font) TTF_CloseFont(font);
	if (initTTF) TTF_Quit();

}

SDL_Surface *Dialog::getSurface() {
	return surf;
}

void Dialog::getRects(vector<SDL_Rect> &list) {
	se.getRects(list);
}

void Dialog::update() {
	for (unsigned int i = 0; i < texts.size(); i++)
		texts[i]->getIds(&se);
	for (unsigned int i = 0; i < buttons.size(); i++)
		buttons[i]->getIds(&se);

	int tractorId[4];
	if (type==CHOOSE) {
		for (int i = 0; i < 4; i++)
			tractorId[i] = se.getNextId();
		SDL_Rect rect;
		rect.x = 88;
		rect.y = 161-tractor[current]->h;
		se.addSprite(tractorId[current],new Sprite(tractor[current],rect,false));
	} else if (type==RESULT) {
		for (unsigned int i = 0; i < stats.size(); i++) {
			SDL_Rect rect;
			SDL_Surface *img = tractor[stats[i].first];
			rect.x = 160-img->w;
			rect.y = 52+49*i;
			se.addSprite(se.getNextId(),new Sprite(img,rect,false));
		}
	}

	for (unsigned int i = 0; i < texts.size(); i++)
		texts[i]->addSprites(&se);
	for (unsigned int i = 0; i < buttons.size(); i++)
		buttons[i]->addSprites(&se);

	se.calcUpdates();

	for (unsigned int i = 0; i < texts.size(); i++)
		texts[i]->updateSprites(&se);
	for (unsigned int i = 0; i < buttons.size(); i++)
		buttons[i]->updateSprites(&se);

	se.update(surf,dialogBkg);

	for (unsigned int i = 0; i < texts.size(); i++)
		texts[i]->doneDrawing(&se);
	for (unsigned int i = 0; i < buttons.size(); i++)
		buttons[i]->doneDrawing(&se);

	se.reset();
}

int Dialog::getButton(int x, int y) {
	for (unsigned int i = 0; i < buttons.size(); i++) {
		if (buttons[i]->isClickable(x,y)) return i;
	}
	return -1;
}

void Dialog::action(int button) {
	switch (type) {
	case QUIT:
	case NEWGAME:
		result = button;
		quit = true;
		break;
	case PLAYERS:
		result = button;
//		if (result>0) result += 1;
		quit = true;
		break;
	case CHOOSE:
		if (button==-1) {
			result = 0;
			quit = true;
		} else {
			if (button==0) {
				result = Game::addPlayer(result,current);
				if (Game::getPlayers(result)==data) quit = true;
				button = 2;
			}
			if (!quit) {
				while (true) {
					int pos = 0;
					for (int i = 0; i < 4; i++)
						if (tractors[i]==current) pos = i;

					pos += (button==2)?1:-1;
					if (pos<0) pos = 3;
					if (pos>3) pos = 0;

					current = tractors[pos];

					if (!Game::hasPlayer(result,current)) break;
				}
			}
		}
		break;
	case RESULT:
		quit = true;
		break;

	}
}

bool Dialog::isClickable(int button) {
	if (button<0) return false;
	switch (type) {
	case CHOOSE:
		if (button==0) return true;
		return Game::getPlayers(result) < 3;
		break;
	default:
		return true;
		break;
	}
}

void Dialog::handleEvent(SDL_Event &event) {
	if (event.type==SDL_MOUSEMOTION) {
		int button = getButton(event.motion.x,event.motion.y);
		if (buttonover>=0 && buttonover!=button) buttons[buttonover]->state = 0;
		buttonover = button;
		bool clickable = isClickable(button);
		if (buttonover>=0) {
			if (clickable)
				buttons[buttonover]->state = buttondown==buttonover?2:1;
			else
				buttons[buttonover]->state = 0;
		}
		if (buttondown>=0) {
			if (clickable)
				wrapper->setPointer(buttondown==buttonover?2:0);
			else
				wrapper->setPointer(0);
		} else {
			wrapper->setPointer(clickable?1:0);
		}
	} else if (event.type==SDL_MOUSEBUTTONDOWN) {
		int button = getButton(event.button.x,event.button.y);
		buttondown = button;
		if (buttondown>=0 && isClickable(buttondown)) {
			buttons[buttondown]->state = 2;
			wrapper->setPointer(2);
		}
	} else if (event.type==SDL_MOUSEBUTTONUP) {
		if (buttondown>=0) {
			int newstate = buttonover==buttondown?1:0;
			if (!isClickable(buttonover)) newstate = 0;
			buttons[buttondown]->state = newstate;
			wrapper->setPointer(newstate);
			if (newstate==1) action(buttondown);
		}
		buttondown = -1;
	} else if (event.type==SDL_KEYDOWN) {
		if (event.key.keysym.sym==SDLK_ESCAPE) {
			switch (type) {
			case QUIT:
			case PLAYERS:
			case RESULT:
				action(0);
				break;
			case NEWGAME:
				action(1);
				break;
			case CHOOSE:
				action(-1);
				break;
			}
		}
	}
}

void Dialog::setWrapper(GameWrapper *w) {
	wrapper = w;
}

bool Dialog::finished() {
	return quit;
}

int Dialog::getResult() {
	return result;
}



